/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"

static void create_neighbor_group(
    MPI_Comm comm_cart_2d, MPI_Group *group_nbr);

INT PX(local_size_ghostcells_internal)(
    const INT *localArraySize, INT tupleSize,
    const INT *numGhostCellsBelow, const INT *numGhostCellsAbove
    )
{
  INT memoryAmount = tupleSize;
  for(int t = 0; t < 3; t++)
    memoryAmount *= numGhostCellsBelow[t] + localArraySize[t] + numGhostCellsAbove[t];
  
  return memoryAmount;
}


PX(gcplan) PX(plan_ghostcells_internal)(
    const INT *n, const PX(permutation) *perm,
    INT tuple_size, const INT *block_col, const INT *block_row,
    const INT *gcells_below, const INT *gcells_above,
    R *data, MPI_Comm comm_col, MPI_Comm comm_row, MPI_Comm comm_cart_2d
    )
{
  int dims = 1, periods = 1, reorder = 0;
  int myrank, rank_prec, rank_next;
  INT loc_ngc_total, local_n[3], local_n_start[3];
  PX(gcplan) ths = NULL;

  MPI_Comm_rank(comm_cart_2d, &myrank);

  /* return NULL for zero gcells or wrong data distribution */
  if( (gcells_below[perm->n[0]] == 0) && (gcells_above[perm->n[0]] == 0) )
    if( (gcells_below[perm->n[1]] == 0) && (gcells_above[perm->n[1]] == 0) )
      if( (gcells_below[perm->n[2]] == 0) && (gcells_above[perm->n[2]] == 0) )
        return NULL;
  
  PX(decompose_3d)(n, perm, block_col, block_row, comm_col, comm_row, local_n, local_n_start);

  ths = (gcplan_internal*) X(malloc)(sizeof(gcplan_internal));

  /* The RMA ghostcell implementation only supports data distributions where the first two dimensions
     are distributed among a two-dimensional processor grid. */
  ths->alg_flag = PFFT_GC_SENDRECV;
  if( (gcells_below[perm->n[2]] == 0) && (gcells_above[perm->n[2]] == 0) )
    if(perm->np[2] == 2)
      ths->alg_flag = PFFT_GC_RMA;

  ths->data = data;
  for(int t=0; t<3; t++){
    ths->n[t]        = n[perm->n[t]];
    ths->loc_n[t]    = local_n[perm->n[t]];
    ths->gc_below[t] = gcells_below[perm->n[t]];
    ths->gc_above[t] = gcells_above[perm->n[t]];
    ths->ngc[t]      = ths->gc_below[t] + ths->loc_n[t] + ths->gc_above[t];

    switch(perm->np[t]){
      case 0:
        ths->blk[t] = block_col[perm->n[t]];
        MPI_Cart_shift(comm_cart_2d, 0, 1, &rank_prec, &rank_next);
        MPI_Comm_dup(comm_col, &ths->comm_cart_1d[t]);
        break;
      case 1:
        ths->blk[t] = block_row[perm->n[t]];
        MPI_Cart_shift(comm_cart_2d, 1, 1, &rank_prec, &rank_next);
        MPI_Comm_dup(comm_row, &ths->comm_cart_1d[t]);
        break;
      case 2:
        ths->blk[t] = ths->n[t];
        rank_prec = rank_next = myrank;
        MPI_Cart_create(MPI_COMM_SELF, 1, &dims, &periods, reorder, &ths->comm_cart_1d[t]);
    }

    if(t == 0){
      ths->rank_left = rank_prec;
      ths->rank_right = rank_next;
    } else if(t == 1){
      ths->rank_lower = rank_prec;
      ths->rank_upper = rank_next;
    } else {
      ths->rank_back = rank_prec;
      ths->rank_front = rank_next;
    }
  }
  ths->tuple = tuple_size;

  for(int t=0; t<3; t++){
    ths->ngc_left[t]  = ths->ngc_right[t] = ths->ngc[t];
    ths->ngc_lower[t] = ths->ngc_upper[t] = ths->ngc[t];
    ths->ngc_back[t]  = ths->ngc_front[t] = ths->ngc[t];
  }
  
  ths->ngc_left[0]  += PX(local_block_size_shifted)(ths->n[0], ths->blk[0], -1, ths->comm_cart_1d[0]) - ths->loc_n[0];
  ths->ngc_right[0] += PX(local_block_size_shifted)(ths->n[0], ths->blk[0], +1, ths->comm_cart_1d[0]) - ths->loc_n[0];
  ths->ngc_lower[1] += PX(local_block_size_shifted)(ths->n[1], ths->blk[1], -1, ths->comm_cart_1d[1]) - ths->loc_n[1];
  ths->ngc_upper[1] += PX(local_block_size_shifted)(ths->n[1], ths->blk[1], +1, ths->comm_cart_1d[1]) - ths->loc_n[1];
  ths->ngc_back[2]  += PX(local_block_size_shifted)(ths->n[2], ths->blk[2], -1, ths->comm_cart_1d[2]) - ths->loc_n[2];
  ths->ngc_front[2] += PX(local_block_size_shifted)(ths->n[2], ths->blk[2], +1, ths->comm_cart_1d[2]) - ths->loc_n[2];
  
  loc_ngc_total = tuple_size;
  for(int t=0; t<3; t++)
    loc_ngc_total *= ths->ngc[t];

  MPI_Win_create(data, loc_ngc_total*sizeof(R), sizeof(R), MPI_INFO_NULL, comm_cart_2d, &(ths->win));
  create_neighbor_group(comm_cart_2d, &(ths->grp));
  
  return ths;
}


static void create_neighbor_group(
    MPI_Comm comm_cart_2d, MPI_Group *group_nbr
    )
{
  int ranks_nbr[4], n_nbr = 0;
  int mpi_left, mpi_right, mpi_upper, mpi_lower;
  MPI_Group group_cart_2d;
  
  MPI_Comm_group(comm_cart_2d, &group_cart_2d);
  
  MPI_Cart_shift(comm_cart_2d, 0, 1, &mpi_left, &mpi_right);
  MPI_Cart_shift(comm_cart_2d, 1, 1, &mpi_lower, &mpi_upper);
  
  ranks_nbr[n_nbr++] = mpi_left;
  if(mpi_right != mpi_left)
    ranks_nbr[n_nbr++] = mpi_right;
  
  ranks_nbr[n_nbr++] = mpi_upper;
  if(mpi_lower != mpi_upper)
    ranks_nbr[n_nbr++] = mpi_lower;
  
  MPI_Group_incl(group_cart_2d, n_nbr, ranks_nbr, group_nbr);
  
  MPI_Group_free(&group_cart_2d);
}


void PX(destroy_gcplan)(
    PX(gcplan) ths
    )
{
  /* plan was already destroyed or never initialized */
  if(ths==NULL)
    return;

  /* take care of MPI variables */
  MPI_Win_free( &(ths->win) );
  MPI_Group_free( &(ths->grp) );
  for(int t=0; t<3; t++)
    MPI_Comm_free( &(ths->comm_cart_1d[t]) );

  /* free memory */
  X(free)(ths);
  /* serfft=NULL; would be senseless, since we can not change the pointer itself */
}


void PX(exchange_ghostcells_internal)(
    PX(gcplan) ths
    )
{
  if(ths == NULL)
    return;

  PX(pad_array_with_zeros_outside)(
      ths->loc_n, ths->gc_below, ths->gc_above, ths->tuple,
      ths->data);

  if(ths->alg_flag & PFFT_GC_RMA)
    PX(exchange_ghostcells_internal_RMA)(ths);
  else
    PX(exchange_ghostcells_internal_sendrecv)(ths);
}


void PX(reduce_ghostcells_internal)(
    PX(gcplan) ths
    )
{
  if(ths == NULL)
    return;

  if(ths->alg_flag & PFFT_GC_RMA)
    PX(reduce_ghostcells_internal_RMA)(ths);
  else
    PX(reduce_ghostcells_internal_sendrecv)(ths);

  PX(truncate_array_outside)(
      ths->ngc, ths->gc_below, ths->gc_above, ths->tuple,
      ths->data);
}
